/*
    Enchanced Regular User Privileges (erup)
    idea by Lukasz Wojtow <lw@wszia.edu.pl>
    code by Rafal Wijata http://www.wijata.com
*/
#include <linux/config.h>
#ifndef CONFIG_ERUP
#include <erup_conf.h>
#endif
#include <erup.h>

#ifndef MODULE
#error "THIS IS MODULE CODE"
#endif

/* we'll modify syscalls */
extern void *sys_call_table[];
#ifdef ERUP_SETID
asmlinkage long (*org_setuid)(uid_t uid);
asmlinkage long (*org_setgid)(gid_t gid);
#ifdef HAVE32ID
asmlinkage long (*org_setuid16)(old_uid_t uid);
asmlinkage long (*org_setgid16)(old_gid_t gid);
#endif
#endif

#ifdef ERUP_SETGROUPS
asmlinkage long (*org_setgroups)(int gidsetsize, gid_t *grouplist);
#ifdef HAVE32ID
asmlinkage long (*org_setgroups16)(int gidsetsize, old_gid_t *grouplist);
#endif
#endif

#ifdef ERUP_CHROOT
asmlinkage long (*org_chroot)(const char *filename);
#endif

#ifdef ERUP_QUOTA
asmlinkage long (*org_quotactl)(int cmd, const char *special, int id, caddr_t addr);
#endif

extern struct file_operations erup_fops;
/***************************** sys functions ************************/
/* we need wrapper for 16bit ids
   it's called with old [g]libc/programm and 2.4+ kernel
   or programms compiled with <2.4 kernels */
#ifdef ERUP_SETID
/* erup_setuid - new version of sys_setuid */
asmlinkage long erup_setuid(id_t uid) {
int	ret = org_setuid(uid);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_uid(current->euid, uid)) {
	    /* grant setuid, no caps becouse of suid */
	    current->uid = current->suid = uid;
	    /* call again, that system will fix caps */
	    ret = org_setuid(uid);
	};
    };
    return ret;
};

/* erup_setgid - new version of sys_setgid */
asmlinkage long erup_setgid(gid_t gid) {
int	ret = org_setgid(gid);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_gid(current->euid, gid)) {
	    /* grant setgid, no caps becouse of sgid */
	    current->gid = current->sgid = gid;
	    /* call again, that system will fix caps */
	    ret = org_setgid(gid);
	};
    };
    return ret;
};
#ifdef HAVE32ID
asmlinkage long erup_setuid16(old_id_t uid) { return erup_setuid((id_t)uid); };
asmlinkage long erup_setgid16(old_id_t gid) { return erup_setgid((id_t)gid); };
#endif

#endif	/* ERUP_SETID */

#ifdef ERUP_SETGROUPS
/* erup_setgroups - new version of sys_setgroups */
asmlinkage long erup_setgroups(int gidsetsize, gid_t *grouplist) {
int	ret = org_setgroups(gidsetsize, grouplist);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_groups(current->euid, gidsetsize)
	    && !cap_raised(current->cap_effective, CAP_SETGID)) {
	    /* grant setgroups */
	    cap_raise(current->cap_effective, CAP_SETGID);
	    /* call again, that system do it its way */
	    ret = org_setgroups(gidsetsize, grouplist);
	    /* revoke setgroups */
	    cap_lower(current->cap_effective, CAP_SETGID);
	};
    };
    return ret;
};
#ifdef HAVE32ID
asmlinkage long erup_setgroups16(int gidsetsize, old_gid_t *grouplist) {
int	ret = org_setgroups16(gidsetsize, grouplist);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_groups(current->euid, gidsetsize)
	    && !cap_raised(current->cap_effective, CAP_SETGID)) {
	    cap_raise(current->cap_effective, CAP_SETGID);
	    ret = org_setgroups16(gidsetsize, grouplist);
	    cap_lower(current->cap_effective, CAP_SETGID);
	};
    };
    return ret;
};
#endif
#endif

#ifdef ERUP_CHROOT
/* erup_chroot - new version of sys_chroot */
asmlinkage long erup_chroot(const char *filename) {
int	ret = org_chroot(filename);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_chr(current->euid, filename, NULL)
	    && !cap_raised(current->cap_effective, CAP_SYS_CHROOT)) {
	    /* grant chroot
	       see include/linux/sched.h & include/linux/capability.h */
	    cap_raise(current->cap_effective, CAP_SYS_CHROOT);
	    /* call again, that system do it its way */
	    ret = org_chroot(filename);
	    /* revoke chroot */
	    cap_lower(current->cap_effective, CAP_SYS_CHROOT);
	};
    };
    return ret;
};
#endif

#ifdef ERUP_QUOTA
/* erup_quotactl - new version of sys_quotactl */
asmlinkage long erup_quotactl(unsigned int cmd, const char *special, int id, caddr_t addr) {
int	ret = org_quotactl(cmd, special, id, addr);	/* call oryginal */
    if (ret) {			/* if failed */
	if (is_allowed_quota(current->euid, cmd >> SUBCMDSHIFT, id)
	    && !cap_raised(current->cap_effective, CAP_SYS_ADMIN)
	    && !cap_raised(current->cap_effective, CAP_SYS_RESOURCE)) {
	    /* grant sys_admin */
	    cap_raise(current->cap_effective, CAP_SYS_ADMIN);	/* 2.4.X */
	    cap_raise(current->cap_effective, CAP_SYS_RESOURCE);/* 2.2.X */
	    /* call again, that system will do it its way */
	    ret = org_quotactl(cmd, special, id, addr);
	    /* revoke sys_admin */
	    cap_lower(current->cap_effective, CAP_SYS_RESOURCE);
	    cap_lower(current->cap_effective, CAP_SYS_ADMIN);
	};
    };
    return ret;
};
#endif
/******************** module code ***********************************/

int init_module(void) {
int	ret;
	if ((ret=erup_init())) return ret;

	lock_kernel();

#ifdef ERUP_SETID
#ifdef HAVE32ID
	org_setuid = sys_call_table[__NR_setuid32];
	sys_call_table[__NR_setuid32] = erup_setuid;
	org_setgid = sys_call_table[__NR_setgid32];
	sys_call_table[__NR_setgid32] = erup_setgid;
	org_setuid16 = sys_call_table[__NR_setuid];
	sys_call_table[__NR_setuid] = erup_setuid16;
	org_setgid16 = sys_call_table[__NR_setgid];
	sys_call_table[__NR_setgid] = erup_setgid16;
#else
	org_setuid = sys_call_table[__NR_setuid];
	sys_call_table[__NR_setuid] = erup_setuid;
	org_setgid = sys_call_table[__NR_setgid];
	sys_call_table[__NR_setgid] = erup_setgid;
#endif
#endif

#ifdef ERUP_SETGROUPS
#ifdef HAVE32ID
	org_setgroups = sys_call_table[__NR_setgroups32];
	sys_call_table[__NR_setgroups32] = erup_setgroups;
	org_setgroups16 = sys_call_table[__NR_setgroups];
	sys_call_table[__NR_setgroups] = erup_setgroups16;
#else
	org_setgroups = sys_call_table[__NR_setgroups];
	sys_call_table[__NR_setgroups] = erup_setgroups;
#endif
#endif

#ifdef ERUP_CHROOT
	org_chroot = sys_call_table[__NR_chroot];
	sys_call_table[__NR_chroot] = erup_chroot;
#endif

#ifdef ERUP_QUOTA
	org_quotactl = sys_call_table[__NR_quotactl];
	sys_call_table[__NR_quotactl] = erup_quotactl;
#endif
	unlock_kernel();

	return NOERR;
};

void cleanup_module(void)
{
	lock_kernel();

#ifdef ERUP_SETID
#ifdef HAVE32ID
	sys_call_table[__NR_setuid32] = org_setuid;
	sys_call_table[__NR_setgid32] = org_setgid;
	sys_call_table[__NR_setuid] = org_setuid16;
	sys_call_table[__NR_setgid] = org_setgid16;
#else
	sys_call_table[__NR_setuid] = org_setuid;
	sys_call_table[__NR_setgid] = org_setgid;
#endif
#endif

#ifdef ERUP_SETGROUPS
#ifdef HAVE32ID
	sys_call_table[__NR_setgroups32] = org_setgroups;
	sys_call_table[__NR_setgroups] = org_setgroups16;
#else
	sys_call_table[__NR_setgroups] = org_setgroups;
#endif
#endif

#ifdef ERUP_CHROOT
	sys_call_table[__NR_chroot] = org_chroot;
#endif

#ifdef ERUP_QUOTA
	sys_call_table[__NR_quotactl] = org_quotactl;
#endif	
	unlock_kernel();

	if (erup_finish()) {
	    /* ?? */
	};
};

MODULE_DESCRIPTION("Enchanced Regular User Privileges");
MODULE_AUTHOR("Rafal Wijata <http://wijata.com>");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
